@@ -15,15 +15,6 @@
 #include <string>                       // for string, operator+, etc
 #include <utility>                      // for pair
 
-#include "iwyu_globals.h"
-#include "iwyu_location_util.h"
-#include "iwyu_path_util.h"
-#include "iwyu_port.h"  // for CHECK_
-#include "iwyu_stl_util.h"
-#include "iwyu_verrs.h"
-#include "llvm/ADT/ArrayRef.h"
-#include "llvm/Support/Casting.h"
-#include "llvm/Support/raw_ostream.h"
 #include "clang/AST/ASTContext.h"
 #include "clang/AST/ASTDumper.h"
 #include "clang/AST/CanonicalType.h"
@@ -31,26 +22,45 @@
 #include "clang/AST/DeclBase.h"
 #include "clang/AST/DeclCXX.h"
 #include "clang/AST/DeclTemplate.h"
+#include "clang/AST/Expr.h"
 #include "clang/AST/ExprCXX.h"
 #include "clang/AST/NestedNameSpecifier.h"
+#include "clang/AST/OperationKinds.h"
+#include "clang/AST/PrettyPrinter.h"
 #include "clang/AST/RecursiveASTVisitor.h"
+#include "clang/AST/Redeclarable.h"
 #include "clang/AST/Stmt.h"
 #include "clang/AST/TemplateBase.h"
 #include "clang/AST/TemplateName.h"
 #include "clang/AST/Type.h"
 #include "clang/AST/TypeLoc.h"
 #include "clang/Basic/Builtins.h"
+#include "clang/Basic/FileEntry.h"
+#include "clang/Basic/IdentifierTable.h"
 #include "clang/Basic/SourceLocation.h"
 #include "clang/Basic/SourceManager.h"
+#include "clang/Basic/Specifiers.h"
+#include "iwyu_globals.h"
+#include "iwyu_location_util.h"
+#include "iwyu_path_util.h"
+#include "iwyu_port.h"  // for CHECK_
+#include "iwyu_stl_util.h"
+#include "iwyu_verrs.h"
+#include "llvm/ADT/ArrayRef.h"
+#include "llvm/ADT/PointerUnion.h"
+#include "llvm/ADT/StringRef.h"
+#include "llvm/Support/Casting.h"
 
-namespace clang {
-class FileEntry;
-}  // namespace clang
+// TODO: Clean out pragmas as IWYU improves.
+// IWYU pragma: no_include "clang/AST/StmtIterator.h"
+// IWYU pragma: no_include "clang/Basic/CustomizableOptional.h"
 
 using clang::ASTDumper;
+using clang::ArrayType;
 using clang::BlockPointerType;
 using clang::CXXConstructExpr;
 using clang::CXXConstructorDecl;
+using clang::CXXDefaultArgExpr;
 using clang::CXXDeleteExpr;
 using clang::CXXDependentScopeMemberExpr;
 using clang::CXXDestructorDecl;
@@ -70,29 +80,39 @@ using clang::DependentScopeDeclRefExpr;
 using clang::DependentTemplateName;
 using clang::DependentTemplateSpecializationType;
 using clang::ElaboratedType;
+using clang::ElaboratedTypeKeyword;
 using clang::EnumDecl;
+using clang::EnumType;
+using clang::ExplicitCastExpr;
 using clang::Expr;
 using clang::ExprWithCleanups;
-using clang::FileEntry;
-using clang::FullSourceLoc;
 using clang::FunctionDecl;
+using clang::FunctionTemplateDecl;
+using clang::FunctionTemplateSpecializationInfo;
 using clang::FunctionType;
+using clang::IdentifierInfo;
 using clang::ImplicitCastExpr;
 using clang::InjectedClassNameType;
 using clang::LValueReferenceType;
 using clang::MemberExpr;
 using clang::MemberPointerType;
 using clang::NamedDecl;
+using clang::NamespaceDecl;
 using clang::NestedNameSpecifier;
 using clang::ObjCObjectType;
+using clang::OptionalFileEntryRef;
 using clang::OverloadExpr;
 using clang::PointerType;
+using clang::PrintingPolicy;
 using clang::QualType;
 using clang::QualifiedTemplateName;
 using clang::RecordDecl;
 using clang::RecordType;
 using clang::RecursiveASTVisitor;
+using clang::Redeclarable;
+using clang::ReferenceType;
 using clang::SourceLocation;
+using clang::SourceManager;
 using clang::SourceRange;
 using clang::Stmt;
 using clang::SubstTemplateTypeParmType;
@@ -107,6 +127,7 @@ using clang::TemplateName;
 using clang::TemplateParameterList;
 using clang::TemplateSpecializationKind;
 using clang::TemplateSpecializationType;
+using clang::TemplateTypeParmDecl;
 using clang::TranslationUnitDecl;
 using clang::Type;
 using clang::TypeAliasTemplateDecl;
@@ -115,6 +136,7 @@ using clang::TypeLoc;
 using clang::TypedefNameDecl;
 using clang::TypedefType;
 using clang::UnaryOperator;
+using clang::UnresolvedLookupExpr;
 using clang::UsingDirectiveDecl;
 using clang::ValueDecl;
 using clang::VarDecl;
@@ -126,6 +148,8 @@ using llvm::dyn_cast_or_null;
 using llvm::errs;
 using llvm::isa;
 using llvm::raw_string_ostream;
+using std::function;
+using std::vector;
 
 namespace include_what_you_use {
 
@@ -181,6 +205,11 @@ static const Type* DesugarAliasTypes(const Type* type) {
   return type;
 }
 
+static bool IsImplicitConversionCtor(const CXXConstructorDecl* ctor) {
+  return !ctor->isExplicit() && ctor->getNumParams() == 1 &&
+         !ctor->isCopyOrMoveConstructor();
+}
+
 }  // anonymous namespace
 
 //------------------------------------------------------------
@@ -200,12 +229,11 @@ SourceLocation ASTNode::GetLocation() const {
   // locations are in a different file, then we're uncertain of our
   // own location.  Return an invalid location.
   if (retval.isValid()) {
-    clang::SourceManager& sm = *GlobalSourceManager();
-    FullSourceLoc full_loc(retval, sm);
-    const FileEntry* spelling_file =
-        sm.getFileEntryForID(sm.getFileID(full_loc.getSpellingLoc()));
-    const FileEntry* instantiation_file =
-        sm.getFileEntryForID(sm.getFileID(full_loc.getExpansionLoc()));
+    SourceManager& sm = *GlobalSourceManager();
+    OptionalFileEntryRef spelling_file =
+        sm.getFileEntryRefForID(sm.getFileID(sm.getSpellingLoc(retval)));
+    OptionalFileEntryRef instantiation_file =
+        sm.getFileEntryRefForID(sm.getFileID(sm.getExpansionLoc(retval)));
     if (spelling_file != instantiation_file)
       return SourceLocation();
   }
@@ -246,7 +274,8 @@ bool IsElaboratedTypeSpecifier(const ASTNode* ast_node) {
   if (ast_node == nullptr)
     return false;
   const ElaboratedType* elaborated_type = ast_node->GetAs<ElaboratedType>();
-  return elaborated_type && elaborated_type->getKeyword() != clang::ETK_None;
+  return elaborated_type &&
+         elaborated_type->getKeyword() != ElaboratedTypeKeyword::None;
 }
 
 const ASTNode* MostElaboratedAncestor(const ASTNode* ast_node) {
@@ -300,16 +329,6 @@ UseFlags ComputeUseFlags(const ASTNode* ast_node) {
   if (IsNodeInsideCXXMethodBody(ast_node))
     flags |= UF_InCxxMethodBody;
 
-  // Definitions of free functions are a little special, because they themselves
-  // count as uses of all prior declarations (ideally we should probably just
-  // require one but it's hard to say which, so we pick all previously seen).
-  // Later IWYU analysis phases do some canonicalization that isn't
-  // necessary/valid for this case, so mark it up for later.
-  if (const auto* fd = ast_node->GetAs<FunctionDecl>()) {
-    if (fd->getKind() == Decl::Function && fd->isThisDeclarationADefinition())
-      flags |= UF_FunctionDfn;
-  }
-
   return flags;
 }
 
@@ -355,6 +374,18 @@ bool IsCXXConstructExprInNewExpr(const ASTNode* ast_node) {
   return ast_node->ParentIsA<CXXNewExpr>();
 }
 
+bool IsAutocastExpr(const ASTNode* ast_node) {
+  const auto* expr = ast_node->GetAs<CXXConstructExpr>();
+  if (!expr)
+    return false;
+  // Explicitly written CXXTemporaryObjectExpr aren't 'autocast' expressions.
+  if (expr->getStmtClass() != Stmt::StmtClass::CXXConstructExprClass)
+    return false;
+  if (!IsImplicitConversionCtor(expr->getConstructor()))
+    return false;
+  return ast_node->template HasAncestorOfType<CallExpr>();
+}
+
 template<typename T>
 NestedNameSpecifier* TryGetQualifier(const ASTNode* ast_node) {
   if (ast_node->IsA<T>())
@@ -410,7 +441,7 @@ string PrintableDecl(const Decl* decl, bool terse/*=true*/) {
     return "<null decl>";
 
   // Use the terse flag to limit the level of output to one line.
-  clang::PrintingPolicy policy = decl->getASTContext().getPrintingPolicy();
+  PrintingPolicy policy = decl->getASTContext().getPrintingPolicy();
   policy.TerseOutput = terse;
   policy.SuppressInitializers = terse;
   policy.PolishForDeclaration = terse;
@@ -504,7 +535,7 @@ void PrintStmt(const Stmt* stmt) {
 string GetWrittenQualifiedNameAsString(const NamedDecl* named_decl) {
   std::string retval;
   llvm::raw_string_ostream ostream(retval);
-  clang::PrintingPolicy printing_policy =
+  PrintingPolicy printing_policy =
       named_decl->getASTContext().getPrintingPolicy();
   printing_policy.SuppressUnwrittenScope = true;
   named_decl->printQualifiedName(ostream, printing_policy);
@@ -522,9 +553,25 @@ static const Type* GetTemplateArgAsType(const TemplateArgument& tpl_arg) {
   return nullptr;
 }
 
-// These utilities figure out the template arguments that are
-// specified in various contexts: TemplateSpecializationType (for
-// template classes) and FunctionDecl (for template functions).
+static bool IsSubstTemplateTypeParmType(const Type* type) {
+  const Type* prev_type = type;
+  do {
+    if (isa<SubstTemplateTypeParmType>(type))
+      return true;
+    if (isa<TypedefType>(type) || isa<TemplateSpecializationType>(type)) {
+      // Still may be substituted outer template parameter but needs additional
+      // analysis outside this function.
+      return false;
+    }
+    prev_type = type;
+    type = type->getLocallyUnqualifiedSingleStepDesugaredType().getTypePtr();
+  } while (type != prev_type);
+  return false;
+}
+
+// These utilities figure out the template arguments and other type components
+// that are specified in various contexts: TemplateSpecializationType (for
+// template classes), FunctionDecl (for template functions), and type aliases.
 //
 // For classes, we care only about explicitly specified template args,
 // not implicit, default args.  For functions, we care about all
@@ -540,10 +587,10 @@ static const Type* GetTemplateArgAsType(const TemplateArgument& tpl_arg) {
 // int(const MyClass&), int, const MyClass&, MyClass).  Note that
 // this function only returns types-as-written, so it does *not* return
 // alloc<int(*)(const MyClass&)>, even though it's part of vector.
-class TypeEnumerator : public RecursiveASTVisitor<TypeEnumerator> {
- public:
-  typedef RecursiveASTVisitor<TypeEnumerator> Base;
 
+class SugaredTypeEnumerator
+    : public RecursiveASTVisitor<SugaredTypeEnumerator> {
+ public:
   // --- Public interface
   // We can add more entry points as needed.
   set<const Type*> Enumerate(const Type* type) {
@@ -574,6 +621,8 @@ class TypeEnumerator : public RecursiveASTVisitor<TypeEnumerator> {
   }
 
  private:
+  typedef RecursiveASTVisitor<SugaredTypeEnumerator> Base;
+
   bool TraverseTypeHelper(QualType qual_type) {
     CHECK_(!qual_type.isNull());
 
@@ -587,7 +636,90 @@ class TypeEnumerator : public RecursiveASTVisitor<TypeEnumerator> {
     // TODO(bolshakov): it doesn't always work properly. For example, both A and
     // TypedefForA would be inserted in seen_types_ for Tpl<A, TypedefForA>,
     // leading to indeterminate resugar_map construction.
-    // TODO(bolshakov): check whether typedef components are "provided".
+    return Base::TraverseType(QualType(desugared, 0));
+  }
+
+  set<const Type*> seen_types_;
+};
+
+class TypeEnumeratorWithoutSubstituted
+    : public RecursiveASTVisitor<TypeEnumeratorWithoutSubstituted> {
+ public:
+  // --- Public interface
+  set<const Type*> Enumerate(const Type* type) {
+    seen_types_.clear();
+    if (!type)
+      return seen_types_;
+    TraverseType(QualType(type, 0));
+    return seen_types_;
+  }
+
+  // --- Methods on RecursiveASTVisitor
+  bool TraverseType(QualType type) {
+    if (type.isNull())
+      return Base::TraverseType(type);
+    return TraverseTypeHelper(type);
+  }
+
+  bool TraverseTypeLoc(TypeLoc type_loc) {
+    if (!type_loc)
+      return Base::TraverseTypeLoc(type_loc);
+    return TraverseTypeHelper(type_loc.getType());
+  }
+
+ private:
+  typedef RecursiveASTVisitor<TypeEnumeratorWithoutSubstituted> Base;
+
+  bool TraverseTypeHelper(QualType qual_type) {
+    CHECK_(!qual_type.isNull());
+    const Type* type = qual_type.getTypePtr();
+    if (IsSubstTemplateTypeParmType(type))
+      return true;
+
+    seen_types_.insert(Desugar(type));
+
+    return Base::TraverseType(qual_type);
+  }
+
+  set<const Type*> seen_types_;
+};
+
+class CanonicalTypeEnumerator
+    : public RecursiveASTVisitor<CanonicalTypeEnumerator> {
+ public:
+  // --- Public interface
+  set<const Type*> Enumerate(const Type* type) {
+    seen_types_.clear();
+    if (!type)
+      return seen_types_;
+    TraverseType(QualType(type, 0));
+    return seen_types_;
+  }
+
+  // --- Methods on RecursiveASTVisitor
+  bool TraverseType(QualType type) {
+    if (type.isNull())
+      return Base::TraverseType(type);
+    return TraverseTypeHelper(type);
+  }
+
+  bool TraverseTypeLoc(TypeLoc type_loc) {
+    if (!type_loc)
+      return Base::TraverseTypeLoc(type_loc);
+    return TraverseTypeHelper(type_loc.getType());
+  }
+
+ private:
+  typedef RecursiveASTVisitor<CanonicalTypeEnumerator> Base;
+
+  bool TraverseTypeHelper(QualType qual_type) {
+    CHECK_(!qual_type.isNull());
+
+    const Type* type = qual_type.getTypePtr();
+    seen_types_.insert(GetCanonicalType(type));
+
+    const Type* desugared = DesugarAliasTypes(type);
+    // Add desugared type components.
     return Base::TraverseType(QualType(desugared, 0));
   }
 
@@ -598,7 +730,17 @@ class TypeEnumerator : public RecursiveASTVisitor<TypeEnumerator> {
 // So 'Foo*' has component 'Foo', as does 'vector<Foo>', while
 // vector<pair<Foo, Bar>> has components pair<Foo,Bar>, Foo, and Bar.
 set<const Type*> GetComponentsOfType(const Type* type) {
-  TypeEnumerator type_enumerator;
+  SugaredTypeEnumerator type_enumerator;
+  return type_enumerator.Enumerate(type);
+}
+
+set<const Type*> GetComponentsOfTypeWithoutSubstituted(const Type* type) {
+  TypeEnumeratorWithoutSubstituted type_enumerator;
+  return type_enumerator.Enumerate(type);
+}
+
+set<const Type*> GetCanonicalComponentsOfType(const Type* type) {
+  CanonicalTypeEnumerator type_enumerator;
   return type_enumerator.Enumerate(type);
 }
 
@@ -609,10 +751,22 @@ bool IsTemplatizedFunctionDecl(const FunctionDecl* decl) {
 }
 
 bool HasImplicitConversionCtor(const CXXRecordDecl* cxx_class) {
+  // Clang leaves ClassTemplateSpecializationDecl empty for uninstantiated
+  // specializations. Hence, replace cxx_class with template definition so that
+  // IWYU can find constructors.
+  if (const auto* tpl_spec =
+          dyn_cast<ClassTemplateSpecializationDecl>(cxx_class)) {
+    if (!tpl_spec->isExplicitSpecialization()) {
+      cxx_class = tpl_spec->getSpecializedTemplate()
+                      ->getCanonicalDecl()
+                      ->getTemplatedDecl();
+    }
+    // TODO(bolshakov): handle partial specializations.
+  }
+
   for (CXXRecordDecl::ctor_iterator ctor = cxx_class->ctor_begin();
        ctor != cxx_class->ctor_end(); ++ctor) {
-    if (ctor->isExplicit() || ctor->getNumParams() != 1 ||
-        ctor->isCopyConstructor() || ctor->isMoveConstructor())
+    if (!IsImplicitConversionCtor(*ctor))
       continue;
     return true;
   }
@@ -635,7 +789,8 @@ bool HasCovariantReturnType(const CXXMethodDecl* method_decl) {
     // of where return type differs is when they're actually covariant.
     // That is, if Clang can already compile this code without errors, and
     // return types differ, it can only be due to covariance.
-    if ((*it)->getReturnType() != derived_return_type)
+    if ((*it)->getReturnType().getCanonicalType() !=
+        derived_return_type.getCanonicalType())
       return true;
   }
 
@@ -738,7 +893,7 @@ static map<const Type*, const Type*> ResugarTypeComponents(
   return retval;
 }
 
-// Helpers for GetTplTypeResugarMapForFunction().
+// Helpers for GetTplInstDataForFunction().
 static map<const Type*, const Type*> GetTplTypeResugarMapForFunctionNoCallExpr(
     const FunctionDecl* decl, unsigned start_arg) {
   map<const Type*, const Type*> retval;
@@ -759,7 +914,6 @@ static map<const Type*, const Type*> GetTplTypeResugarMapForFunctionNoCallExpr(
 
 static map<const Type*, const Type*>
 GetTplTypeResugarMapForFunctionExplicitTplArgs(
-    const FunctionDecl* decl,
     const TemplateArgumentListInfo& explicit_tpl_list) {
   map<const Type*, const Type*> retval;
   for (const TemplateArgumentLoc& loc : explicit_tpl_list.arguments()) {
@@ -810,9 +964,41 @@ static const Type* GetSugaredTypeOf(const Expr* expr) {
   return v.sugared.getTypePtr();
 }
 
-map<const Type*, const Type*> GetTplTypeResugarMapForFunction(
-    const FunctionDecl* decl, const Expr* calling_expr) {
-  map<const Type*, const Type*> retval;
+static map<const Type*, const Type*> GetDefaultedArgResugarMap(
+    const FunctionDecl* decl) {
+  map<const Type*, const Type*> res;
+  const FunctionTemplateDecl* template_decl = decl->getPrimaryTemplate();
+  if (!template_decl)
+    return res;
+  const TemplateParameterList* params = template_decl->getTemplateParameters();
+  const TemplateArgumentList* args = decl->getTemplateSpecializationArgs();
+  const unsigned count = params->size();
+  CHECK_(args->size() == count);
+  for (unsigned i = 0; i < count; ++i) {
+    if (const auto* param_decl =
+            dyn_cast<TemplateTypeParmDecl>(params->getParam(i))) {
+      if (param_decl->isParameterPack())
+        continue;
+      if (!param_decl->hasDefaultArgument())
+        continue;
+      const QualType arg_type = args->get(i).getAsType().getCanonicalType();
+      const QualType default_arg_type = param_decl->getDefaultArgument()
+                                            .getArgument()
+                                            .getAsType()
+                                            .getCanonicalType();
+      if (arg_type == default_arg_type) {
+        res.emplace(arg_type.getTypePtr(), nullptr);
+      }
+    }
+  }
+  return res;
+}
+
+TemplateInstantiationData GetTplInstDataForFunction(
+    const FunctionDecl* decl, const Expr* calling_expr,
+    function<set<const Type*>(const Type*)> provided_getter) {
+  map<const Type*, const Type*> resugar_map;
+  set<const Type*> provided_types;
 
   // If calling_expr is nullptr, then we can't find any explicit template
   // arguments, if they were specified (e.g. 'Fn<int>()'), and we
@@ -820,9 +1006,10 @@ map<const Type*, const Type*> GetTplTypeResugarMapForFunction(
   // can't resugar at all.  We just have to hope that the types happen
   // to be already sugared, because the actual-type is already canonical.
   if (calling_expr == nullptr) {
-    retval = GetTplTypeResugarMapForFunctionNoCallExpr(decl, 0);
-    retval = ResugarTypeComponents(retval);  // add in retval's decomposition
-    return retval;
+    resugar_map = GetTplTypeResugarMapForFunctionNoCallExpr(decl, 0);
+    resugar_map = ResugarTypeComponents(
+        resugar_map);  // add in resugar_map's decomposition
+    return TemplateInstantiationData{resugar_map, {}};
   }
 
   // If calling_expr is a CXXConstructExpr of CXXNewExpr, then it's
@@ -844,8 +1031,8 @@ map<const Type*, const Type*> GetTplTypeResugarMapForFunction(
     const TemplateArgumentListInfo& explicit_tpl_args =
         GetExplicitTplArgs(callee_expr);
     if (explicit_tpl_args.size() > 0) {
-      retval = GetTplTypeResugarMapForFunctionExplicitTplArgs(
-          decl, explicit_tpl_args);
+      resugar_map =
+          GetTplTypeResugarMapForFunctionExplicitTplArgs(explicit_tpl_args);
       start_of_implicit_args = explicit_tpl_args.size();
     }
   } else {
@@ -853,11 +1040,14 @@ map<const Type*, const Type*> GetTplTypeResugarMapForFunction(
     const TemplateArgumentListInfo& explicit_tpl_args =
         GetExplicitTplArgs(calling_expr);
     if (explicit_tpl_args.size() > 0) {
-      retval = GetTplTypeResugarMapForFunctionExplicitTplArgs(
-          decl, explicit_tpl_args);
-      retval = ResugarTypeComponents(retval);
+      resugar_map =
+          GetTplTypeResugarMapForFunctionExplicitTplArgs(explicit_tpl_args);
+      resugar_map = ResugarTypeComponents(resugar_map);
+      for (const auto [_, sugared_type] : resugar_map)
+        InsertAllInto(provided_getter(sugared_type), &provided_types);
     }
-    return retval;
+    InsertAllInto(GetDefaultedArgResugarMap(decl), &resugar_map);
+    return TemplateInstantiationData{resugar_map, provided_types};
   }
 
   // Now we have to figure out, as best we can, the sugar-mappings for
@@ -879,37 +1069,51 @@ map<const Type*, const Type*> GetTplTypeResugarMapForFunction(
   //                 under it, take the pre-cast type instead?
   set<const Type*> fn_arg_types;
   for (unsigned i = 0; i < num_args; ++i) {
+    if (isa<CXXDefaultArgExpr>(fn_args[i]))
+      break;
     const Type* argtype = GetSugaredTypeOf(fn_args[i]);
     // TODO(csilvers): handle RecordTypes that are a TemplateSpecializationDecl
     InsertAllInto(GetComponentsOfType(argtype), &fn_arg_types);
   }
 
   for (const Type* type : fn_arg_types) {
-    // See if any of the template args in retval are the desugared form of us.
+    // See if any of the template args in resugar_map are the desugared form
+    // of us.
     const Type* desugared_type = GetCanonicalType(type);
     if (ContainsKey(desugared_types, desugared_type)) {
-      retval[desugared_type] = type;
+      resugar_map[desugared_type] = type;
       if (desugared_type != type) {
         VERRS(6) << "Remapping template arg of interest: "
                  << PrintableType(desugared_type) << " -> "
                  << PrintableType(type) << "\n";
       }
     }
+    // TODO(bolshakov): more fine-grained determination of provided types could
+    // e.g. consider only arguments that actually take part in type deduction.
+    // Simply moving the line below into the if-statement body above doesn't
+    // work because 'type' can be just an internal component of a 'typedef'ed
+    // argument.
+    InsertAllInto(provided_getter(type), &provided_types);
   }
 
+  InsertAllInto(GetDefaultedArgResugarMap(decl), &resugar_map);
+
   // Log the types we never mapped.
   for (const auto& types : desugared_types) {
-    if (!ContainsKey(retval, types.first)) {
+    if (!ContainsKey(resugar_map, types.first)) {
       VERRS(6) << "Ignoring unseen-in-fn-args template arg of interest: "
                << PrintableType(types.first) << "\n";
     }
   }
 
-  retval = ResugarTypeComponents(retval);  // add in the decomposition of retval
-  return retval;
+  resugar_map = ResugarTypeComponents(
+      resugar_map);  // add in the decomposition of resugar_map
+  for (const auto [_, sugared_type] : resugar_map)
+    InsertAllInto(provided_getter(sugared_type), &provided_types);
+  return TemplateInstantiationData{resugar_map, provided_types};
 }
 
-const NamedDecl* GetInstantiatedFromDecl(const CXXRecordDecl* class_decl) {
+const NamedDecl* GetInstantiatedFromDecl(const NamedDecl* class_decl) {
   if (const ClassTemplateSpecializationDecl* tpl_sp_decl =
       DynCastFrom(class_decl)) {  // an instantiated class template
     PointerUnion<ClassTemplateDecl*, ClassTemplatePartialSpecializationDecl*>
@@ -969,12 +1173,21 @@ bool IsFriendDecl(const Decl* decl) {
   return decl->getFriendObjectKind() != Decl::FOK_None;
 }
 
-bool IsExplicitInstantiation(const clang::Decl* decl) {
+bool IsExplicitInstantiation(const Decl* decl) {
   TemplateSpecializationKind kind = GetTemplateSpecializationKind(decl);
   return kind == clang::TSK_ExplicitInstantiationDeclaration ||
          kind == clang::TSK_ExplicitInstantiationDefinition;
 }
 
+bool IsExplicitInstantiationDefinitionAsWritten(
+    const ClassTemplateSpecializationDecl* decl) {
+  // When swithing instantiation declaration to definition, clang preserves
+  // the 'extern' keyword location info.
+  return decl->getSpecializationKind() ==
+             clang::TSK_ExplicitInstantiationDefinition &&
+         decl->getExternKeywordLoc().isInvalid();
+}
+
 bool IsInInlineNamespace(const Decl* decl) {
   const DeclContext* dc = decl->getDeclContext();
   for (; dc; dc = dc->getParent()) {
@@ -985,6 +1198,17 @@ bool IsInInlineNamespace(const Decl* decl) {
   return false;
 }
 
+bool IsInNamespace(const NamedDecl* decl, const NamespaceDecl* ns_decl) {
+  const DeclContext* primary_ns_context = ns_decl->getPrimaryContext();
+  for (const DeclContext* dc = decl->getDeclContext(); dc;
+       dc = dc->getParent()) {
+    if (dc->getPrimaryContext() == primary_ns_context)
+      return true;
+  }
+
+  return false;
+}
+
 bool IsForwardDecl(const NamedDecl* decl) {
   if (const auto* tag_decl = dyn_cast<TagDecl>(decl)) {
     // clang-format off
@@ -1013,16 +1237,16 @@ bool HasDefaultTemplateParameters(const TemplateDecl* decl) {
   return tpl_params->getMinRequiredArguments() < tpl_params->size();
 }
 
-template <class T> inline set<const clang::NamedDecl*> GetRedeclsOfRedeclarable(
-    const clang::Redeclarable<T>* decl) {
-  return set<const clang::NamedDecl*>(decl->redecls_begin(),
-                                      decl->redecls_end());
+template <class T>
+inline set<const NamedDecl*> GetRedeclsOfRedeclarable(
+    const Redeclarable<T>* decl) {
+  return set<const NamedDecl*>(decl->redecls_begin(), decl->redecls_end());
 }
 
 // The only way to find out whether a decl can be dyn_cast to a
 // Redeclarable<T> and what T is is to enumerate the possibilities.
 // Hence we hard-code the list.
-set<const clang::NamedDecl*> GetNonTagRedecls(const clang::NamedDecl* decl) {
+set<const NamedDecl*> GetNonTagRedecls(const NamedDecl* decl) {
   CHECK_(!isa<TagDecl>(decl) && "For tag types, call GetTagRedecls()");
   CHECK_(!isa<ClassTemplateDecl>(decl) && "For tpls, call GetTagRedecls()");
   // TODO(wan): go through iwyu to replace TypedefDecl with
@@ -1034,7 +1258,7 @@ set<const clang::NamedDecl*> GetNonTagRedecls(const clang::NamedDecl* decl) {
   if (const VarDecl* specific_decl = DynCastFrom(decl))
     return GetRedeclsOfRedeclarable(specific_decl);
   // Not redeclarable, so the output is just the input.
-  set<const clang::NamedDecl*> retval;
+  set<const NamedDecl*> retval;
   retval.insert(decl);
   return retval;
 }
@@ -1073,14 +1297,16 @@ set<const NamedDecl*> GetTagRedecls(const NamedDecl* decl) {
 
 const NamedDecl* GetFirstRedecl(const NamedDecl* decl) {
   const NamedDecl* first_decl = decl;
-  FullSourceLoc first_decl_loc(GetLocation(first_decl), *GlobalSourceManager());
+  SourceLocation first_decl_loc = GetLocation(first_decl);
+
   set<const NamedDecl*> all_redecls = GetTagRedecls(decl);
   if (all_redecls.empty())  // input is not a class or class template
     return nullptr;
 
+  SourceManager& sm = *GlobalSourceManager();
   for (const NamedDecl* redecl : all_redecls) {
-    const FullSourceLoc redecl_loc(GetLocation(redecl), *GlobalSourceManager());
-    if (redecl_loc.isBeforeInTranslationUnitThan(first_decl_loc)) {
+    SourceLocation redecl_loc = GetLocation(redecl);
+    if (sm.isBeforeInTranslationUnit(redecl_loc, first_decl_loc)) {
       first_decl = redecl;
       first_decl_loc = redecl_loc;
     }
@@ -1110,8 +1336,8 @@ bool DeclsAreInSameClass(const Decl* decl1, const Decl* decl2) {
   return decl1->getDeclContext()->isRecord();
 }
 
-bool IsBuiltinFunction(const clang::NamedDecl* decl) {
-  if (const clang::IdentifierInfo* iden = decl->getIdentifier()) {
+bool IsBuiltinFunction(const NamedDecl* decl) {
+  if (const IdentifierInfo* iden = decl->getIdentifier()) {
     unsigned builtin_id = iden->getBuiltinID();
     if (builtin_id != 0) {
       const clang::Builtin::Context& ctx = decl->getASTContext().BuiltinInfo;
@@ -1122,15 +1348,35 @@ bool IsBuiltinFunction(const clang::NamedDecl* decl) {
   return false;
 }
 
+bool IsImplicitlyInstantiatedDfn(const FunctionDecl* decl) {
+  const FunctionTemplateSpecializationInfo* tpl_spec_info =
+      decl->getTemplateSpecializationInfo();
+  if (!tpl_spec_info)
+    return false;  // Not a template specialization.
+  return decl->isThisDeclarationADefinition() &&
+         tpl_spec_info->getTemplateSpecializationKind() ==
+             clang::TSK_ImplicitInstantiation;
+}
+
+const CXXMethodDecl* GetFromLeastDerived(const CXXMethodDecl* decl) {
+  while (decl->size_overridden_methods())
+    decl = *decl->begin_overridden_methods();
+  return decl;
+}
+
 // --- Utilities for Type.
 
 const Type* GetTypeOf(const Expr* expr) {
+  if (const auto* decl_ref_expr = dyn_cast<DeclRefExpr>(expr))
+    return decl_ref_expr->getDecl()->getType().getTypePtr();
+  if (const auto* cast_expr = dyn_cast<ExplicitCastExpr>(expr))
+    return cast_expr->getTypeAsWritten().getTypePtr();
   return expr->getType().getTypePtr();
 }
 
 const Type* GetTypeOf(const CXXConstructExpr* expr) {
   const Type* type = expr->getType().getTypePtr();
-  if (const clang::ArrayType* array_type = type->getAsArrayTypeUnsafe()) {
+  if (const ArrayType* array_type = type->getAsArrayTypeUnsafe()) {
     type = array_type->getElementType().getTypePtr();
   }
   return type;
@@ -1184,17 +1430,10 @@ const Type* Desugar(const Type* type) {
 }
 
 bool IsTemplatizedType(const Type* type) {
-  return (type && isa<TemplateSpecializationType>(Desugar(type)));
+  return type && type->getAs<TemplateSpecializationType>();
 }
 
-bool IsClassType(const clang::Type* type) {
-  type = Desugar(type);
-  return (type &&
-          (isa<TemplateSpecializationType>(type) || isa<RecordType>(type)));
-}
-
-bool InvolvesTypeForWhich(const Type* type,
-                          std::function<bool(const Type*)> pred) {
+bool InvolvesTypeForWhich(const Type* type, function<bool(const Type*)> pred) {
   type = Desugar(type);
   if (pred(type))
     return true;
@@ -1215,7 +1454,7 @@ bool InvolvesTypeForWhich(const Type* type,
 
 bool IsPointerOrReferenceAsWritten(const Type* type) {
   type = Desugar(type);
-  return isa<PointerType>(type) || isa<LValueReferenceType>(type);
+  return isa<PointerType>(type) || isa<ReferenceType>(type);
 }
 
 const Type* RemovePointersAndReferencesAsWritten(const Type* type) {
@@ -1301,6 +1540,12 @@ const NamedDecl* TypeToDeclForContent(const Type* type) {
   return TypeToDeclImpl(type, /*as_written=*/false);
 }
 
+const Type* RemoveReference(const Type* type) {
+  if (const auto* ref_type = type->getAs<ReferenceType>())
+    return ref_type->getPointeeType().getTypePtr();
+  return type;
+}
+
 const Type* RemoveReferenceAsWritten(const Type* type) {
   if (const LValueReferenceType* ref_type = DynCastFrom(type))
     return ref_type->getPointeeType().getTypePtr();
@@ -1328,22 +1573,18 @@ bool HasImplicitConversionConstructor(const Type* type) {
   return HasImplicitConversionCtor(cxx_class);
 }
 
-map<const clang::Type*, const clang::Type*>
-GetTplTypeResugarMapForClassNoComponentTypes(const clang::Type* type) {
+static TemplateInstantiationData GetTplInstDataForClassNoComponentTypes(
+    ArrayRef<TemplateArgument> written_tpl_args,
+    const ClassTemplateSpecializationDecl* cls_tpl_decl,
+    function<set<const Type*>(const Type*)> provided_getter) {
   map<const Type*, const Type*> retval;
-  const auto* tpl_spec_type = type->getAs<TemplateSpecializationType>();
-  if (!tpl_spec_type) {
-    return retval;
-  }
 
   // Pull the template arguments out of the specialization type. If this is
   // a ClassTemplateSpecializationDecl specifically, we want to
   // get the arguments therefrom to correctly handle default arguments.
-  llvm::ArrayRef<TemplateArgument> tpl_args = tpl_spec_type->template_arguments();
+  llvm::ArrayRef<TemplateArgument> tpl_args = written_tpl_args;
   unsigned num_args = tpl_args.size();
 
-  const NamedDecl* decl = TypeToDeclAsWritten(tpl_spec_type);
-  const auto* cls_tpl_decl = dyn_cast<ClassTemplateSpecializationDecl>(decl);
   if (cls_tpl_decl) {
     const TemplateArgumentList& tpl_arg_list =
         cls_tpl_decl->getTemplateInstantiationArgs();
@@ -1356,10 +1597,11 @@ GetTplTypeResugarMapForClassNoComponentTypes(const clang::Type* type) {
   // explicitly specified type may fulfill multiple template args:
   //   template <typename R, typename A1> struct Foo<R(A1)> { ... }
   set<unsigned> explicit_args;   // indices into tpl_args we've filled
-  TypeEnumerator type_enumerator;
-  for (unsigned i = 0; i < tpl_spec_type->template_arguments().size(); ++i) {
+  SugaredTypeEnumerator type_enumerator;
+  set<const Type*> provided_types;
+  for (unsigned i = 0; i < written_tpl_args.size(); ++i) {
     set<const Type*> arg_components =
-        type_enumerator.Enumerate(tpl_spec_type->template_arguments()[i]);
+        type_enumerator.Enumerate(written_tpl_args[i]);
     // Go through all template types mentioned in the arg-as-written,
     // and compare it against each of the types in the template decl
     // (the latter are all desugared).  If there's a match, update
@@ -1377,6 +1619,9 @@ GetTplTypeResugarMapForClassNoComponentTypes(const clang::Type* type) {
         }
       }
     }
+
+    for (const Type* component : arg_components)
+      InsertAllInto(provided_getter(component), &provided_types);
   }
 
   // Now take a look at the args that were not filled explicitly.
@@ -1390,19 +1635,82 @@ GetTplTypeResugarMapForClassNoComponentTypes(const clang::Type* type) {
     }
   }
 
-  return retval;
+  return TemplateInstantiationData{retval, provided_types};
 }
 
-map<const clang::Type*, const clang::Type*> GetTplTypeResugarMapForClass(
-    const clang::Type* type) {
-  return ResugarTypeComponents(  // add in the decomposition of retval
-      GetTplTypeResugarMapForClassNoComponentTypes(type));
+TemplateInstantiationData GetTplInstDataForClassNoComponentTypes(
+    const clang::Type* type,
+    std::function<set<const clang::Type*>(const clang::Type*)>
+        provided_getter) {
+  const auto* tpl_spec_type = type->getAs<TemplateSpecializationType>();
+  if (!tpl_spec_type)
+    return TemplateInstantiationData{};
+  const NamedDecl* decl = TypeToDeclAsWritten(tpl_spec_type);
+  const auto* cls_tpl_decl = dyn_cast<ClassTemplateSpecializationDecl>(decl);
+  return GetTplInstDataForClassNoComponentTypes(
+      tpl_spec_type->template_arguments(), cls_tpl_decl, provided_getter);
 }
 
-bool CanBeOpaqueDeclared(const clang::EnumType* type) {
+TemplateInstantiationData GetTplInstDataForClass(
+    const Type* type, function<set<const Type*>(const Type*)> provided_getter) {
+  TemplateInstantiationData result =
+      GetTplInstDataForClassNoComponentTypes(type, provided_getter);
+  InsertAllInto(provided_getter(type), &result.provided_types);
+  return TemplateInstantiationData{
+      ResugarTypeComponents(
+          result.resugar_map),  // add in the decomposition of retval
+      result.provided_types};
+}
+
+TemplateInstantiationData GetTplInstDataForClass(
+    ArrayRef<TemplateArgument> written_tpl_args,
+    const ClassTemplateSpecializationDecl* cls_tpl_decl,
+    function<set<const Type*>(const Type*)> provided_getter) {
+  TemplateInstantiationData result = GetTplInstDataForClassNoComponentTypes(
+      written_tpl_args, cls_tpl_decl, provided_getter);
+  return TemplateInstantiationData{
+      ResugarTypeComponents(
+          result.resugar_map),  // add in the decomposition of retval
+      result.provided_types};
+}
+
+bool CanBeOpaqueDeclared(const EnumType* type) {
   return type->getDecl()->isFixed();
 }
 
+vector<const Type*> GetCanonicalArgComponents(
+    const TemplateSpecializationType* type) {
+  vector<const Type*> res;
+  SugaredTypeEnumerator enumerator;
+  for (const TemplateArgument& arg : type->template_arguments()) {
+    for (const Type* component : enumerator.Enumerate(arg))
+      res.push_back(GetCanonicalType(component));
+  }
+  return res;
+}
+
+bool IsReferenceToModifiableLValue(const Type* type) {
+  return type->isLValueReferenceType() &&
+         !type->getPointeeType().getQualifiers().hasConst();
+}
+
+bool IsDerivedToBasePtrConvertible(const Type* derived_ptr_type,
+                                   const Type* base_ptr_type) {
+  if (!derived_ptr_type->isPointerType() || !base_ptr_type->isPointerType())
+    return false;
+
+  const QualType derived = derived_ptr_type->getPointeeType();
+  const QualType base = base_ptr_type->getPointeeType();
+  if (const CXXRecordDecl* derived_decl = derived->getAsCXXRecordDecl()) {
+    if (const CXXRecordDecl* base_decl = base->getAsCXXRecordDecl()) {
+      if (!derived_decl->isDerivedFrom(base_decl))
+        return false;
+      return base.isAtLeastAsQualifiedAs(derived, base_decl->getASTContext());
+    }
+  }
+  return false;
+}
+
 // --- Utilities for Stmt.
 
 bool IsAddressOf(const Expr* expr) {
@@ -1411,6 +1719,13 @@ bool IsAddressOf(const Expr* expr) {
   return false;
 }
 
+bool IsDependentNameCall(const Expr* expr) {
+  const CallExpr* call_expr = dyn_cast<CallExpr>(expr);
+  if (!call_expr)
+    return false;
+  return isa<UnresolvedLookupExpr>(call_expr->getCallee());
+}
+
 const Type* TypeOfParentIfMethod(const CallExpr* expr) {
   // callee_expr is a MemberExpr if we're a normal class method, or
   // DeclRefExpr if we're a static class method or an overloaded operator.
@@ -1433,7 +1748,7 @@ const Expr* GetFirstClassArgument(CallExpr* expr) {
       // If a method is called, return 'this'.
       return expr->getArg(0);
     }
-    // Handle free functions.
+    // Handle non-member functions.
     CHECK_(callee_decl->getNumParams() == expr->getNumArgs() &&
         "Require one-to-one match between call arguments and decl parameters");
     int params_count = callee_decl->getNumParams();
@@ -1531,4 +1846,9 @@ bool IsDeclaredInsideFunction(const Decl* decl) {
   return isa<FunctionDecl>(decl_ctx);
 }
 
+bool IsDeclaredInsideMacro(const Decl* decl) {
+  SourceRange range = decl->getSourceRange();
+  return range.getBegin().isMacroID() || range.getEnd().isMacroID();
+}
+
 }  // namespace include_what_you_use
