@@ -12,20 +12,20 @@
 #ifndef INCLUDE_WHAT_YOU_USE_IWYU_AST_UTIL_H_
 #define INCLUDE_WHAT_YOU_USE_IWYU_AST_UTIL_H_
 
+#include <functional>                   // for function
 #include <map>                          // for map
 #include <set>                          // for set
 #include <string>                       // for string
+#include <vector>                       // for vector
 
-#include "iwyu_port.h"  // for CHECK_
-#include "iwyu_use_flags.h"
-#include "llvm/Support/Casting.h"
-#include "clang/AST/DeclBase.h"
 #include "clang/AST/NestedNameSpecifier.h"
-#include "clang/AST/Stmt.h"
 #include "clang/AST/TemplateBase.h"
 #include "clang/AST/Type.h"
 #include "clang/AST/TypeLoc.h"
 #include "clang/Basic/SourceLocation.h"
+#include "iwyu_port.h"  // for CHECK_
+#include "iwyu_use_flags.h"
+#include "llvm/Support/Casting.h"
 
 namespace clang {
 class CXXConstructExpr;
@@ -35,20 +35,28 @@ class CXXDestructorDecl;
 class CXXMethodDecl;
 class CXXRecordDecl;
 class CallExpr;
-class CastExpr;
 class ClassTemplateDecl;
+class ClassTemplateSpecializationDecl;
+class Decl;
+class DeclContext;
 class Expr;
 class FunctionDecl;
 class NamedDecl;
+class NamespaceDecl;
+class Stmt;
 class TagDecl;
 class TemplateDecl;
 class TemplateName;
 class TranslationUnitDecl;
 class TypeDecl;
 class ValueDecl;
-struct ASTTemplateArgumentListInfo;
 }  // namespace clang
 
+namespace llvm {
+template <typename>
+class ArrayRef;
+}  // namespace llvm
+
 namespace include_what_you_use {
 
 using std::map;
@@ -193,6 +201,17 @@ class ASTNode {
 
   // Convenience methods.
 
+  template <typename To>
+  const To* GetAncestorAs() const {
+    // DynCast needs a dummy argument of type To* to help its resolution.
+    const To* dummy = nullptr;
+    for (const ASTNode* node = this; node != nullptr; node = node->parent_) {
+      if (const auto result = node->DynCast<To>(dummy))
+        return result;
+    }
+    return nullptr;
+  }
+
   template<typename To> bool AncestorIsA(int generation) const {
     return GetAncestorAs<To>(generation) != nullptr;
   }
@@ -431,6 +450,10 @@ bool IsCXXConstructExprInInitializer(const ASTNode* ast_node);
 // Returns true if this node is a CXXConstructExpr in CXXNewExpr.
 bool IsCXXConstructExprInNewExpr(const ASTNode* ast_node);
 
+// Determines if the given node represents implicit one-argument constructor
+// call in context of function argument initialization - so called 'autocast'.
+bool IsAutocastExpr(const ASTNode*);
+
 // If ASTNode is of a kind that has a qualifier (something that
 // comes before the ::), return that, else return nullptr.
 const clang::NestedNameSpecifier* GetQualifier(const ASTNode* ast_node);
@@ -531,6 +554,11 @@ clang::SourceRange GetSourceRangeOfClassDecl(const clang::Decl* decl);
 set<clang::FunctionDecl*> GetLateParsedFunctionDecls(
     clang::TranslationUnitDecl* decl);
 
+struct TemplateInstantiationData {
+  map<const clang::Type*, const clang::Type*> resugar_map;
+  set<const clang::Type*> provided_types;
+};
+
 // One can't have partial template specialization or default template
 // args for function templates, but they're complicated in their own
 // way: they can have deduced template arguments (deduced from the
@@ -563,9 +591,12 @@ set<clang::FunctionDecl*> GetLateParsedFunctionDecls(
 // NOTE: This routine is far from perfect.  To really do this right,
 // we'd need to refactor SemaTemplateDeduction to take an argument to
 // not canonicalize deduced template arguments.
+// Besides, the returned data contains a set of "provided" types (e.g. from
+// a typedef being type template argument).
 // calling_expr should be a CallExpr, CXXConstructExpr, or DeclRefExpr.
-map<const clang::Type*, const clang::Type*> GetTplTypeResugarMapForFunction(
-    const clang::FunctionDecl* decl, const clang::Expr* calling_expr);
+TemplateInstantiationData GetTplInstDataForFunction(
+    const clang::FunctionDecl* decl, const clang::Expr* calling_expr,
+    std::function<set<const clang::Type*>(const clang::Type*)> provided_getter);
 
 // If class_decl is instantiated from a class template,
 // returns the decl for that template; otherwise returns class_decl.
@@ -588,7 +619,7 @@ map<const clang::Type*, const clang::Type*> GetTplTypeResugarMapForFunction(
 // return the template decl, which provides the actual class body.
 // We try to return a decl that's also a definition, when possible.
 const clang::NamedDecl* GetInstantiatedFromDecl(
-    const clang::CXXRecordDecl* class_decl);
+    const clang::NamedDecl* class_decl);
 
 // For an implicitly instantiated templated c++ class -- that is, a
 // class like vector<int> that isn't explicitly written in the source
@@ -608,9 +639,19 @@ bool IsFriendDecl(const clang::Decl* decl);
 // or definition.
 bool IsExplicitInstantiation(const clang::Decl* decl);
 
+// If an explicit instantiation definition follows the corresponding explicit
+// instantiation declaration, clang marks the first declaration as definition.
+// This function returns true if the given specialization declaration
+// corresponds to the genuine instantiation definition, as written
+// in the source.
+bool IsExplicitInstantiationDefinitionAsWritten(
+    const clang::ClassTemplateSpecializationDecl*);
+
 // Returns true if this decl is nested inside an inline namespace.
 bool IsInInlineNamespace(const clang::Decl* decl);
 
+bool IsInNamespace(const clang::NamedDecl*, const clang::NamespaceDecl*);
+
 // Returns true if a named decl looks like a forward-declaration of a
 // class (rather than a definition, a friend declaration, or an 'in
 // place' declaration like 'struct Foo' in 'void MyFunc(struct Foo*);'
@@ -672,6 +713,15 @@ bool DeclsAreInSameClass(const clang::Decl* decl1, const clang::Decl* decl2);
 // Returns true if the given decl/name is a builtin function
 bool IsBuiltinFunction(const clang::NamedDecl* decl);
 
+// Returns true if the function decl is an implicitly instantiated definition
+// (in particular, not just a declaration).
+bool IsImplicitlyInstantiatedDfn(const clang::FunctionDecl*);
+
+// If the given method overrides base class methods, returns the overridden
+// method declaration from the least derived class, otherwise returns the given
+// argument.
+const clang::CXXMethodDecl* GetFromLeastDerived(const clang::CXXMethodDecl*);
+
 // --- Utilities for Type.
 
 const clang::Type* GetTypeOf(const clang::Expr* expr);
@@ -695,14 +745,29 @@ const clang::Type* Desugar(const clang::Type* type);
 // A 'component' of a type is a type beneath it in the AST tree.
 // So 'Foo*' has component 'Foo', as does 'vector<Foo>', while
 // vector<pair<Foo, Bar>> has components pair<Foo,Bar>, Foo, and Bar.
+// All sugar is thrown away except typedefs, type aliases and template
+// specialization types. Given such a typedef:
+// typedef Tpl1<A, B> Alias;
+// the returned set for something like 'Tpl2<Alias>' contains 'Alias', 'A' and
+// 'B' but not 'Tpl1<A, B>'.
 set<const clang::Type*> GetComponentsOfType(const clang::Type* type);
 
+// Almost the same except it returns canonical types.
+set<const clang::Type*> GetCanonicalComponentsOfType(const clang::Type* type);
+
+// Returns types for determination of their "provision" status. They are
+// canonicalized because intermediate sugar should be always provided already
+// according to language rules. Substituted template parameter types (and their
+// underlying types) cannot be provided. E. g., given such a template
+// template <class T> struct Tpl1 { typedef Tpl2<T, A> Alias; };
+// the returned set for the type 'Tpl1<B>::Alias' contains 'A' but not 'B'
+// or any sugar for 'B'.
+set<const clang::Type*> GetComponentsOfTypeWithoutSubstituted(
+    const clang::Type*);
+
 // Returns true if the type has any template arguments.
 bool IsTemplatizedType(const clang::Type* type);
 
-// Returns true if the type is a RecordType or a TemplateSpecializationType.
-bool IsClassType(const clang::Type* type);
-
 // Returns true if any type involved (recursively examining template
 // arguments) satisfies the given predicate.
 bool InvolvesTypeForWhich(const clang::Type* type,
@@ -715,6 +780,10 @@ bool InvolvesTypeForWhich(const clang::Type* type,
 // TODO(csilvers): what about array-type?
 bool IsPointerOrReferenceAsWritten(const clang::Type* type);
 
+// This function removes a reference even it is hidden by some type sugar.
+const clang::Type* RemoveReference(const clang::Type*);
+
+// This function doesn't "see through" type sugar.
 const clang::Type* RemoveReferenceAsWritten(const clang::Type* type);
 
 // Gets rid of all the pointers and references to get to the 'base'
@@ -775,17 +844,26 @@ bool HasImplicitConversionConstructor(const clang::Type* type);
 // We also include mappings for component types: if we have an entry
 // 'vector<TypedefType>' -> 'vector<Foo>', we also add an entry
 // 'TypedefType' -> 'Foo'.
+// Besides, the returned data contains a set of "provided" types (e.g. from
+// a typedef being type template argument).
 // For ease of calling, this accept any type, but will return an empty
-// map for any input that's not a template specialization type.
-map<const clang::Type*, const clang::Type*> GetTplTypeResugarMapForClass(
-    const clang::Type* type);
+// result for any input that's not a template specialization type.
+TemplateInstantiationData GetTplInstDataForClass(
+    const clang::Type* type,
+    std::function<set<const clang::Type*>(const clang::Type*)> provided_getter);
 
-// Like GetTplTypeResugarMapForClass, but if a type has
+TemplateInstantiationData GetTplInstDataForClass(
+    llvm::ArrayRef<clang::TemplateArgument> written_tpl_args,
+    const clang::ClassTemplateSpecializationDecl* cls_tpl_decl,
+    std::function<set<const clang::Type*>(const clang::Type*)> provided_getter);
+
+// Like GetTplInstDataForClass, but if a type has
 // components (for instance, 'Foo*' and 'vector<Foo>' both
 // have a component Foo), we don't include the components
 // in the result-map.
-map<const clang::Type*, const clang::Type*>
-GetTplTypeResugarMapForClassNoComponentTypes(const clang::Type* type);
+TemplateInstantiationData GetTplInstDataForClassNoComponentTypes(
+    const clang::Type* type,
+    std::function<set<const clang::Type*>(const clang::Type*)> provided_getter);
 
 // Returns true if, for the given enumeration type, opaque (i.e. forward,
 // in fact) declarations are allowed. It means that the enumeration should be
@@ -793,11 +871,28 @@ GetTplTypeResugarMapForClassNoComponentTypes(const clang::Type* type);
 // according to the standard.
 bool CanBeOpaqueDeclared(const clang::EnumType* type);
 
+// Collects template argument type components and returns them desugared.
+// The result may contain duplicates.
+std::vector<const clang::Type*> GetCanonicalArgComponents(
+    const clang::TemplateSpecializationType*);
+
+bool IsReferenceToModifiableLValue(const clang::Type*);
+
+// Returns true when the arguments are pointer types referring to classes or
+// structs related by direct inheritance. Also checks that their qualifiers
+// allow the conversion.
+bool IsDerivedToBasePtrConvertible(const clang::Type* derived_ptr_type,
+                                   const clang::Type* base_ptr_type);
+
 // --- Utilities for Stmt.
 
 // Returns true if the given expr is '&<something>'.
 bool IsAddressOf(const clang::Expr* expr);
 
+// Returns true if the given expr is a call-expression targeting an
+// unresolved/dependent name.
+bool IsDependentNameCall(const clang::Expr* expr);
+
 // If this function call comes from a class method -- either a normal
 // one or a static one -- returns the type of the class.  Otherwise,
 // returns nullptr.  Note that static class methods do *not* have a
@@ -806,7 +901,7 @@ const clang::Type* TypeOfParentIfMethod(const clang::CallExpr* expr);
 
 // Given a function call, return the first argument that's a class
 // (possibly a template specialization).  Note we ignore pointers to a
-// class.  This is used with 'free' overloaded operators ('ostream&
+// class.  This is used with non-member overloaded operators ('ostream&
 // operator<<(ostream& a, int x)' to figure out what class the
 // operator 'logically' belongs to.  This is a heuristic (the operator
 // may "belong" to more than one argument, for instance), but covers
@@ -843,6 +938,9 @@ std::string GetKindName(const clang::TypeLoc typeloc);
 // visible from said function.
 bool IsDeclaredInsideFunction(const clang::Decl* decl);
 
+// Returns true if decl is partially inside a macro.
+bool IsDeclaredInsideMacro(const clang::Decl* decl);
+
 }  // namespace include_what_you_use
 
 #endif  // INCLUDE_WHAT_YOU_USE_IWYU_AST_UTIL_H_
