@@ -12,10 +12,21 @@
 #include <algorithm>                    // for sort, find
 #include <cstdio>                       // for snprintf
 #include <iterator>                     // for inserter
+#include <list>
 #include <map>                          // for _Rb_tree_const_iterator, etc
 #include <utility>                      // for pair, make_pair, operator>
 #include <vector>                       // for vector, vector<>::iterator, etc
 
+#include "clang/AST/ASTContext.h"
+#include "clang/AST/Decl.h"
+#include "clang/AST/DeclBase.h"
+#include "clang/AST/DeclCXX.h"
+#include "clang/AST/DeclTemplate.h"
+#include "clang/AST/DeclarationName.h"
+#include "clang/AST/PrettyPrinter.h"
+#include "clang/AST/Type.h"
+#include "clang/AST/TypeLoc.h"
+#include "clang/Basic/SourceLocation.h"
 #include "iwyu_ast_util.h"
 #include "iwyu_globals.h"
 #include "iwyu_include_picker.h"
@@ -25,37 +36,37 @@
 #include "iwyu_stl_util.h"
 #include "iwyu_string_util.h"
 #include "iwyu_verrs.h"
+#include "llvm/ADT/STLExtras.h"
+#include "llvm/ADT/StringRef.h"
 #include "llvm/Support/Casting.h"
-#include "llvm/Support/raw_ostream.h"
-#include "clang/AST/ASTContext.h"
-#include "clang/AST/Decl.h"
-#include "clang/AST/DeclBase.h"
-#include "clang/AST/DeclTemplate.h"
-#include "clang/AST/Type.h"
-#include "clang/Basic/SourceLocation.h"
 
 namespace include_what_you_use {
 
+using clang::BuiltinTemplateDecl;
+using clang::CXXRecordDecl;
 using clang::ClassTemplateDecl;
 using clang::ClassTemplateSpecializationDecl;
-using clang::CXXMethodDecl;
-using clang::CXXRecordDecl;
 using clang::Decl;
 using clang::DeclContext;
+using clang::DeclarationName;
+using clang::ElaboratedTypeLoc;
 using clang::EnumDecl;
-using clang::FileEntry;
 using clang::FunctionDecl;
 using clang::NamedDecl;
 using clang::NamespaceDecl;
+using clang::OptionalFileEntryRef;
+using clang::PrintingPolicy;
 using clang::RecordDecl;
 using clang::SourceLocation;
 using clang::SourceRange;
 using clang::TagDecl;
 using clang::TemplateDecl;
+using clang::TypeSourceInfo;
 using clang::UsingDecl;
 using llvm::cast;
 using llvm::dyn_cast;
 using llvm::errs;
+using llvm::find_if;
 using llvm::isa;
 using llvm::raw_string_ostream;
 using std::map;
@@ -140,17 +151,16 @@ string OutputLine::printable_line(size_t min_length, size_t max_length) const {
 // to a FakeNamedDecl. When a FakeNamedDecl is created, it will be
 // inserted into the map with itself as the key (implicitly casted to
 // a NamedDecl).
-std::map<const clang::NamedDecl*, const FakeNamedDecl*>
-g_fake_named_decl_map;
+std::map<const NamedDecl*, const FakeNamedDecl*> g_fake_named_decl_map;
 
 // Since dynamic casting is not an option, this method is provided to
 // determine if a decl is actually a FakeNamedDecl.
-const FakeNamedDecl* FakeNamedDeclIfItIsOne(const clang::NamedDecl* decl) {
+const FakeNamedDecl* FakeNamedDeclIfItIsOne(const NamedDecl* decl) {
   return GetOrDefault(g_fake_named_decl_map, decl, nullptr);
 }
 
 std::string PrintableUnderlyingType(const EnumDecl* enum_decl) {
-  if (const clang::TypeSourceInfo* type_source_info =
+  if (const TypeSourceInfo* type_source_info =
           enum_decl->getIntegerTypeSourceInfo()) {
     return " : " + type_source_info->getType().getAsString(
                        enum_decl->getASTContext().getPrintingPolicy());
@@ -163,11 +173,11 @@ std::string PrintableUnderlyingType(const EnumDecl* enum_decl) {
 
 FakeNamedDecl::FakeNamedDecl(const string& kind_name, const string& qual_name,
                              const string& decl_filepath, int decl_linenum)
-    : clang::NamedDecl(clang::Decl::Record, nullptr, clang::SourceLocation(),
-                       clang::DeclarationName()),
+    : NamedDecl(Decl::Record, nullptr, SourceLocation(), DeclarationName()),
       kind_name_(kind_name),
       qual_name_(qual_name),
-      decl_filepath_(decl_filepath), decl_linenum_(decl_linenum) {
+      decl_filepath_(decl_filepath),
+      decl_linenum_(decl_linenum) {
   g_fake_named_decl_map[this] = this;
 }
 
@@ -178,7 +188,7 @@ FakeNamedDecl::FakeNamedDecl(const string& kind_name, const string& qual_name,
 // call ourselves.
 
 string GetKindName(const TagDecl* tag_decl) {
-  const clang::NamedDecl* const named_decl = tag_decl;
+  const NamedDecl* const named_decl = tag_decl;
   if (const FakeNamedDecl* fake = FakeNamedDeclIfItIsOne(named_decl)) {
     return fake->kind_name();
   }
@@ -191,7 +201,7 @@ string GetKindName(const TagDecl* tag_decl) {
   return tag_decl->getKindName().str();
 }
 
-string GetQualifiedNameAsString(const clang::NamedDecl* named_decl) {
+string GetQualifiedNameAsString(const NamedDecl* named_decl) {
   if (const FakeNamedDecl* fake = FakeNamedDeclIfItIsOne(named_decl)) {
     return fake->qual_name();
   }
@@ -199,7 +209,7 @@ string GetQualifiedNameAsString(const clang::NamedDecl* named_decl) {
 }
 
 // Name we put in the comments next to an #include.
-string GetShortNameAsString(const clang::NamedDecl* named_decl) {
+string GetShortNameAsString(const NamedDecl* named_decl) {
   if (const FakeNamedDecl* fake = FakeNamedDeclIfItIsOne(named_decl)) {
     return fake->qual_name();
   }
@@ -267,7 +277,7 @@ OneUse::OneUse(const NamedDecl* decl, SourceLocation use_loc,
       decl_file_(GetFileEntry(decl_loc_)),
       decl_filepath_(GetFilePath(decl_file_)),
       use_loc_(use_loc),
-      use_kind_(use_kind),             // full use or fwd-declare use
+      use_kind_(use_kind),  // full use or fwd-declare use
       use_flags_(flags),
       comment_(comment ? comment : ""),
       ignore_use_(false),
@@ -275,32 +285,46 @@ OneUse::OneUse(const NamedDecl* decl, SourceLocation use_loc,
 }
 
 // This constructor always creates a full use.
-OneUse::OneUse(const string& symbol_name, const FileEntry* dfn_file,
-               const string& dfn_filepath, SourceLocation use_loc)
+OneUse::OneUse(const string& symbol_name, OptionalFileEntryRef dfn_file,
+               SourceLocation use_loc)
     : symbol_name_(symbol_name),
       short_symbol_name_(symbol_name),
       decl_(nullptr),
       decl_file_(dfn_file),
-      decl_filepath_(dfn_filepath),
+      decl_filepath_(GetFilePath(dfn_file)),
       use_loc_(use_loc),
       use_kind_(kFullUse),
       use_flags_(UF_None),
       ignore_use_(false),
       is_iwyu_violation_(false) {
-  // Sometimes dfn_filepath is actually a fully quoted include.  In
-  // that case, we take that as an unchangable mapping that we
-  // should never remove, so we make it the suggested header.
-  CHECK_(!decl_filepath_.empty() && "Must pass a real filepath to OneUse");
-  if (decl_filepath_[0] == '"' || decl_filepath_[0] == '<')
-    suggested_header_ = decl_filepath_;
+  CHECK_(dfn_file && "OneUse: dfn_file must be set");
+  CHECK_(!decl_filepath_.empty() && "OneUse: dfn_file must have a name");
+  CHECK_(!IsQuotedInclude(decl_filepath_))
+      << "OneUse: dfn_file must have a real name, was: " << decl_filepath_;
 }
 
-void OneUse::reset_decl(const clang::NamedDecl* decl) {
-    CHECK_(decl_ && "Need existing decl to reset it");
-    CHECK_(decl && "Need to reset decl with existing decl");
-    decl_ = decl;
-    decl_file_ = GetFileEntry(decl);
-    decl_filepath_ = GetFilePath(decl);
+OneUse::OneUse(OptionalFileEntryRef included_file, const string& quoted_include)
+    : symbol_name_(),
+      short_symbol_name_(),
+      decl_(nullptr),
+      decl_file_(included_file),
+      decl_filepath_(quoted_include),
+      use_loc_(SourceLocation()),
+      use_kind_(kFullUse),
+      use_flags_(UF_None),
+      ignore_use_(false),
+      is_iwyu_violation_(false) {
+  CHECK_(IsQuotedInclude(decl_filepath_))
+      << "OneUse: bad quoted_include: " << quoted_include;
+  suggested_header_ = decl_filepath_;
+}
+
+void OneUse::reset_decl(const NamedDecl* decl) {
+  CHECK_(decl_ && "Need existing decl to reset it");
+  CHECK_(decl && "Need to reset decl with existing decl");
+  decl_ = decl;
+  decl_file_ = GetFileEntry(decl);
+  decl_filepath_ = GetFilePath(decl);
 }
 
 int OneUse::UseLinenum() const {
@@ -315,14 +339,8 @@ void OneUse::SetPublicHeaders() {
   // We should never need to deal with public headers if we already know
   // who we map to.
   CHECK_(suggested_header_.empty() && "Should not need a public header here");
-  const IncludePicker& picker = GlobalIncludePicker();   // short alias
-  const string use_path = GetFilePath(use_loc_);
-  // If the symbol has a special mapping, use it, otherwise map its file.
-  public_headers_ = picker.GetCandidateHeadersForSymbolUsedFrom(
-      symbol_name_, use_path);
-  if (public_headers_.empty())
-    public_headers_ = picker.GetCandidateHeadersForFilepathIncludedFrom(
-        decl_filepath(), use_path);
+  public_headers_ = GlobalIncludePicker().GetMappedPublicHeaders(
+      symbol_name_, GetFilePath(use_loc_), decl_filepath());
   if (public_headers_.empty())
     public_headers_.push_back(ConvertToQuotedInclude(decl_filepath()));
 }
@@ -449,29 +467,32 @@ string MungedForwardDeclareLineForNontemplates(const TagDecl* decl) {
 // forward-declare the template, e.g.
 //     "namespace ns { template <typename T> class Foo; }".
 string MungedForwardDeclareLineForTemplates(const TemplateDecl* decl) {
-  // DeclPrinter prints the class name just as we like it (with
-  // default args and everything) -- with logic that doesn't exist
-  // elsewhere in clang that I can see.  Unfortunately, it also prints
-  // the full class body.  So, as a hack, we use PrintableDecl to get
-  // the full declaration, and then hack off everything after the
-  // template name.  We also have to replace the name with the fully
-  // qualified name.  TODO(csilvers): prepend namespaces instead.
-  std::string line;       // llvm wants regular string, not our versa-string
+  // DeclPrinter prints the class name just as we like it (with default args and
+  // everything) -- with logic that doesn't exist elsewhere in clang that I can
+  // see. Unfortunately, it also prints the full class body. So, as a hack, we
+  // postprocess the printed decl and then cut off everything after the template
+  // name. We also have to replace the name with the fully qualified name.
+  // TODO(csilvers): prepend namespaces instead.
+  std::string line;
   raw_string_ostream ostream(line);
-  decl->print(ostream);   // calls DeclPrinter
-  line = ostream.str();
 
-  // Remove "final" specifier which isn't needed for forward
-  // declarations.
+  // Print the decl using PolishForDeclaration, which strips some semantic
+  // attributes.
+  PrintingPolicy policy = decl->getASTContext().getPrintingPolicy();
+  policy.PolishForDeclaration = true;
+  decl->print(ostream, policy);
+  ostream.flush();
+
+  // Remove "final" specifier, it isn't allowed for forward declarations.
   ReplaceAll(&line, " final ", " ");
 
   // Get rid of the superclasses, if any (this will nix the body too).
-  line = Split(line, " :", 2)[0];
+  DropFrom(&line, " :");
   // Get rid of the template body, if any (true if no superclasses).
-  line = Split(line, " {", 2)[0];
+  DropFrom(&line, " {");
 
-  // The template name is now the last word on the line.  Replace it
-  // by its fully-qualified form.
+  // The template name is now the last word on the line. Replace it by its
+  // fully-qualified form.
   const string::size_type name = line.rfind(' ');
   CHECK_(name != string::npos && "Unexpected printable template-type");
 
@@ -492,13 +513,7 @@ string MungedForwardDeclareLine(const NamedDecl* decl) {
 
 OneIncludeOrForwardDeclareLine::OneIncludeOrForwardDeclareLine(
     const NamedDecl* fwd_decl)
-    : line_(internal::MungedForwardDeclareLine(fwd_decl)),
-      start_linenum_(-1),   // set 'for real' below
-      end_linenum_(-1),     // set 'for real' below
-      is_desired_(false),
-      is_present_(false),
-      included_file_(nullptr),
-      fwd_decl_(fwd_decl) {
+    : line_(internal::MungedForwardDeclareLine(fwd_decl)), fwd_decl_(fwd_decl) {
   const SourceRange decl_lines = GetSourceRangeOfClassDecl(fwd_decl);
   // We always want to use the instantiation line numbers: for code like
   //     FORWARD_DECLARE_CLASS(MyClass);
@@ -508,15 +523,24 @@ OneIncludeOrForwardDeclareLine::OneIncludeOrForwardDeclareLine(
 }
 
 OneIncludeOrForwardDeclareLine::OneIncludeOrForwardDeclareLine(
-    const FileEntry* included_file, const string& quoted_include, int linenum)
+    ElaboratedTypeLoc type_loc)
+    : is_desired_(true),
+      is_present_(true),
+      is_elaborated_type_(true),
+      fwd_decl_(type_loc.getTypePtr()->getOwnedTagDecl()) {
+  const SourceRange decl_lines = type_loc.getLocalSourceRange();
+  start_linenum_ = GetLineNumber(GetInstantiationLoc(decl_lines.getBegin()));
+  end_linenum_ = GetLineNumber(GetInstantiationLoc(decl_lines.getEnd()));
+}
+
+OneIncludeOrForwardDeclareLine::OneIncludeOrForwardDeclareLine(
+    OptionalFileEntryRef included_file, const string& quoted_include,
+    int linenum)
     : line_("#include " + quoted_include),
       start_linenum_(linenum),
       end_linenum_(linenum),
-      is_desired_(false),
-      is_present_(false),
       quoted_include_(quoted_include),
-      included_file_(included_file),
-      fwd_decl_(nullptr) {
+      included_file_(included_file) {
 }
 
 bool OneIncludeOrForwardDeclareLine::HasSymbolUse(const string& symbol_name)
@@ -540,16 +564,16 @@ string OneIncludeOrForwardDeclareLine::LineNumberString() const {
   return buf;
 }
 
-IwyuFileInfo::IwyuFileInfo(const clang::FileEntry* this_file,
+IwyuFileInfo::IwyuFileInfo(OptionalFileEntryRef this_file,
                            const IwyuPreprocessorInfo* preprocessor_info,
                            const string& quoted_include_name)
-  : file_(this_file),
-    preprocessor_info_(preprocessor_info),
-    quoted_file_(quoted_include_name),
-    is_prefix_header_(false),
-    is_pch_in_code_(false),
-    desired_includes_have_been_calculated_(false)
-{}
+    : file_(this_file),
+      preprocessor_info_(preprocessor_info),
+      quoted_file_(quoted_include_name),
+      is_prefix_header_(false),
+      is_pch_in_code_(false),
+      desired_includes_have_been_calculated_(false) {
+}
 
 void IwyuFileInfo::AddAssociatedHeader(const IwyuFileInfo* other) {
   VERRS(6) << "Adding " << GetFilePath(other->file_)
@@ -557,7 +581,7 @@ void IwyuFileInfo::AddAssociatedHeader(const IwyuFileInfo* other) {
   associated_headers_.insert(other);
 }
 
-void IwyuFileInfo::AddInclude(const clang::FileEntry* includee,
+void IwyuFileInfo::AddInclude(OptionalFileEntryRef includee,
                               const string& quoted_includee, int linenumber) {
   OneIncludeOrForwardDeclareLine new_include(includee, quoted_includee,
                                              linenumber);
@@ -585,7 +609,7 @@ void IwyuFileInfo::AddInclude(const clang::FileEntry* includee,
            << " -> " << GetFilePath(includee) << "\n";
 }
 
-void IwyuFileInfo::AddForwardDeclare(const clang::NamedDecl* fwd_decl,
+void IwyuFileInfo::AddForwardDeclare(const NamedDecl* fwd_decl,
                                      bool definitely_keep_fwd_decl) {
   CHECK_(fwd_decl && "forward_declare_decl unexpectedly nullptr");
   CHECK_((isa<ClassTemplateDecl>(fwd_decl) || isa<TagDecl>(fwd_decl)) &&
@@ -601,6 +625,15 @@ void IwyuFileInfo::AddForwardDeclare(const clang::NamedDecl* fwd_decl,
            << internal::GetQualifiedNameAsString(fwd_decl) << "\n";
 }
 
+void IwyuFileInfo::AddElaboratedType(ElaboratedTypeLoc type_loc) {
+  if (!type_loc.getTypePtr()->getOwnedTagDecl())
+    return;
+  lines_.push_back(OneIncludeOrForwardDeclareLine(type_loc));
+  VERRS(6) << "Found owning elaborated type: " << GetFilePath(file_) << ":"
+           << lines_.back().LineNumberString() << ": "
+           << PrintableTypeLoc(type_loc) << "\n";
+}
+
 void IwyuFileInfo::AddUsingDecl(const UsingDecl* using_decl) {
   CHECK_(using_decl && "using_decl unexpectedly nullptr");
   using_decl_referenced_.insert(std::make_pair(using_decl, false));
@@ -636,14 +669,15 @@ void IwyuFileInfo::ReportFullSymbolUse(SourceLocation use_loc,
     const NamedDecl* report_decl;
     SourceLocation report_decl_loc;
 
-    if ((flags & (UF_FunctionDfn | UF_ExplicitInstantiation)) == 0) {
+    if ((flags & (UF_DefinitionUse | UF_ExplicitInstantiation)) == 0) {
       // Since we need the full symbol, we need the decl's definition-site too.
       // Also, by default we canonicalize the location, using GetLocation.
       report_decl = GetDefinitionAsWritten(decl);
       report_decl_loc = GetLocation(report_decl);
     } else {
-      // However, if we're defining the function or we are targeting an explicit
-      // instantiation, we want to use it as-is and not try to canonicalize at all.
+      // However, if a declaration is used by its own definition or we are
+      // targeting an explicit instantiation, we want to use them as-is and not
+      // try to canonicalize at all.
       report_decl = decl;
       report_decl_loc = decl->getLocation();
     }
@@ -655,33 +689,30 @@ void IwyuFileInfo::ReportFullSymbolUse(SourceLocation use_loc,
 }
 
 void IwyuFileInfo::ReportFullSymbolUse(SourceLocation use_loc,
-                                       const FileEntry* dfn_file,
+                                       OptionalFileEntryRef dfn_file,
                                        const string& symbol) {
-  symbol_uses_.push_back(
-      OneUse(symbol, dfn_file, GetFilePath(dfn_file), use_loc));
+  symbol_uses_.push_back(OneUse(symbol, dfn_file, use_loc));
   LogSymbolUse("Marked full-info use of symbol", symbol_uses_.back());
 }
 
-void IwyuFileInfo::ReportMacroUse(clang::SourceLocation use_loc,
-                                  clang::SourceLocation dfn_loc,
+void IwyuFileInfo::ReportMacroUse(SourceLocation use_loc,
+                                  SourceLocation dfn_loc,
                                   const string& symbol) {
-  symbol_uses_.push_back(OneUse(symbol, GetFileEntry(dfn_loc),
-                                GetFilePath(dfn_loc), use_loc));
+  symbol_uses_.push_back(OneUse(symbol, GetFileEntry(dfn_loc), use_loc));
   LogSymbolUse("Marked full-info use of macro", symbol_uses_.back());
 }
 
-void IwyuFileInfo::ReportDefinedMacroUse(const clang::FileEntry* used_in) {
+void IwyuFileInfo::ReportDefinedMacroUse(OptionalFileEntryRef used_in) {
   macro_users_.insert(used_in);
 }
 
-void IwyuFileInfo::ReportIncludeFileUse(const clang::FileEntry* included_file,
+void IwyuFileInfo::ReportIncludeFileUse(OptionalFileEntryRef included_file,
                                         const string& quoted_include) {
-  symbol_uses_.push_back(OneUse("", included_file, quoted_include,
-                                SourceLocation()));
+  symbol_uses_.push_back(OneUse(included_file, quoted_include));
   LogSymbolUse("Marked use of include-file", symbol_uses_.back());
 }
 
-void IwyuFileInfo::ReportKnownDesiredFile(const FileEntry* included_file) {
+void IwyuFileInfo::ReportKnownDesiredFile(OptionalFileEntryRef included_file) {
   kept_includes_.insert(included_file);
 }
 
@@ -797,13 +828,13 @@ bool DeclIsVisibleToUseInSameFile(const Decl* decl, const OneUse& use) {
 // might be "calculate minimal-ish includes". :-)  It populates
 // each OneUse in uses with the best #include for that use.
 // direct_includes: this file's direct includes only.
-// associated_direct_includes: direct includes for 'associated'
+// associated_desired_includes: desired includes for 'associated'
 // files.  For everything but foo.cc, this is empty; for foo.cc it's
 // foo.h's includes and foo-inl.h's includes.
 set<string> CalculateMinimalIncludes(
     const string& use_quoted_include,
     const set<string>& direct_includes,
-    const set<string>& associated_direct_includes,
+    const set<string>& associated_desired_includes,
     vector<OneUse>* uses) {
   set<string> desired_headers;
 
@@ -844,9 +875,8 @@ set<string> CalculateMinimalIncludes(
   // Steps (2): Go through the needed private-includes that map to
   // more than one public #include.  Use the following priority order:
   // - Ourselves.
-  // - An include in associated_direct_includes (those are includes
-  //   that are not going away, since we can't change associated
-  //   files).
+  // - An include in associated_desired_includes (those are includes
+  //   that are not going away, since they should be calculated already).
   // - Includes in direct_includes that are also already in
   //   desired_headers.
   // - Includes in desired_headers.
@@ -869,7 +899,7 @@ set<string> CalculateMinimalIncludes(
     for (const string& choice : public_headers) {
       if (use.has_suggested_header())
         break;
-      if (ContainsKey(associated_direct_includes, choice)) {
+      if (ContainsKey(associated_desired_includes, choice)) {
         use.set_suggested_header(choice);
         desired_headers.insert(use.suggested_header());
         LogIncludeMapping("in associated header", use);
@@ -1013,6 +1043,9 @@ set<string> CalculateMinimalIncludes(
 //     a.h does not see b.h in its transitive #includes.  (Note: This
 //     happens before include-picker mapping, so it's still possible to
 //     see 'new' includes via a manual mapping.)
+// B8) Discard uses of NamespaceDecls where the file also uses symbols
+//     within that namespace. IWYU for the symbol will also satisfy IWYU
+//     for the containing namespace(s).
 // B1') Discard macro uses in the same file as the definition (B3 redux).
 // B2') Discard macro uses that form a 'backwards' #include (B6 redux).
 // B3') Discard macro uses from a 'new' #include (B7 redux).
@@ -1200,8 +1233,8 @@ static bool HasMapping(const string& symbol) {
   return !GlobalIncludePicker().GetCandidateHeadersForSymbol(symbol).empty();
 }
 
-void ProcessFullUse(OneUse* use,
-                    const IwyuPreprocessorInfo* preprocessor_info) {
+void ProcessFullUse(OneUse* use, const IwyuPreprocessorInfo* preprocessor_info,
+                    const vector<OneUse>& all_uses) {
   CHECK_(use->decl() && "Must call ProcessFullUse on a decl");
   CHECK_(use->is_full_use() && "Must not call ProcessFullUse on fwd-decl");
   if (use->ignore_use())   // we're already ignoring it
@@ -1266,10 +1299,11 @@ void ProcessFullUse(OneUse* use,
   // definition from iwyu's point of view.)  We don't bother with
   // RedeclarableTemplate<> types (FunctionTemplateDecl), since for
   // those types, iwyu *does* care about the definition vs declaration.
-  // All this is moot when FunctionDecls are being defined, all their redecls
+  // All this is moot for decl uses triggered by definitions, all their redecls
   // are separately registered as uses so that a definition anchors all its
   // declarations.
-  if (!(use->flags() & UF_FunctionDfn) && !is_builtin_function_with_mappings) {
+  if (!(use->flags() & UF_DefinitionUse) &&
+      !is_builtin_function_with_mappings) {
     set<const NamedDecl*> all_redecls;
     if (isa<TagDecl>(use->decl()) || isa<ClassTemplateDecl>(use->decl()))
       all_redecls.insert(use->decl());  // for classes, just consider the dfn
@@ -1288,7 +1322,7 @@ void ProcessFullUse(OneUse* use,
 
   // (B4) Discard symbol uses for builtin symbols, including new/delete and
   // template builtins.
-  if (isa<clang::BuiltinTemplateDecl>(use->decl())) {
+  if (isa<BuiltinTemplateDecl>(use->decl())) {
     VERRS(6) << "Ignoring use of " << use->symbol_name()
              << " (" << use->PrintableUseLoc() << "): built-in template\n";
     use->set_ignore_use();
@@ -1323,8 +1357,8 @@ void ProcessFullUse(OneUse* use,
     // See if we also recorded a use of the parent.
     const NamedDecl* parent_dfn = GetDefinitionAsWritten(parent_decl);
 
-    const FileEntry* decl_file_entry = GetFileEntry(use->decl_loc());
-    const FileEntry* parent_file_entry =
+    OptionalFileEntryRef decl_file_entry = GetFileEntry(use->decl_loc());
+    OptionalFileEntryRef parent_file_entry =
         GetFileEntry(GetInstantiationLoc(GetLocation(parent_dfn)));
 
     // We want to map the definition-files to their public headers if
@@ -1388,6 +1422,20 @@ void ProcessFullUse(OneUse* use,
       return;
     }
   }
+
+  // (B8) Discard uses of NamespaceDecls where the file also uses
+  // symbols within that namespace. IWYU for the symbol will also
+  // satisfy IWYU for the containing namespace(s).
+  if (const auto* ns_decl = dyn_cast<NamespaceDecl>(use->decl())) {
+    if (std::any_of(
+            all_uses.begin(), all_uses.end(), [ns_decl](const OneUse& use) {
+              const NamedDecl* decl = use.decl();
+              return decl && (ns_decl != decl) && IsInNamespace(decl, ns_decl);
+            })) {
+      use->set_ignore_use();
+      return;
+    }
+  }
 }
 
 void ProcessSymbolUse(OneUse* use,
@@ -1395,7 +1443,7 @@ void ProcessSymbolUse(OneUse* use,
   if (use->ignore_use())   // we're already ignoring it
     return;
 
-  const FileEntry* use_file = GetFileEntry(use->use_loc());
+  OptionalFileEntryRef use_file = GetFileEntry(use->use_loc());
   const string quoted_decl_file = ConvertToQuotedInclude(use->decl_filepath());
 
   // (B1') Like (B2), discard symbol uses in the same file as their definition.
@@ -1433,10 +1481,9 @@ void ProcessSymbolUse(OneUse* use,
 }
 
 void CalculateIwyuForForwardDeclareUse(
-    OneUse* use,
-    const set<string>& actual_includes,
+    OneUse* use, const set<string>& actual_includes,
     const set<string>& desired_includes,
-    const set<const FileEntry*>& associated_includes) {
+    const set<OptionalFileEntryRef>& associated_includes) {
   CHECK_(!use->ignore_use() && "Trying to calculate on an ignored use");
   CHECK_(use->decl() && "CalculateIwyuForForwardDeclareUse takes a fwd-decl");
   CHECK_(!use->is_full_use() && "ForwardDeclareUse are not full uses");
@@ -1629,7 +1676,7 @@ void IwyuFileInfo::CalculateIwyuViolations(vector<OneUse>* uses) {
   }
   for (OneUse& use : *uses) {
     if (use.is_full_use() && use.decl())
-      internal::ProcessFullUse(&use, preprocessor_info_);
+      internal::ProcessFullUse(&use, preprocessor_info_, *uses);
   }
   for (OneUse& use : *uses) {
     if (use.is_full_use() && !use.decl())
@@ -1649,8 +1696,9 @@ void IwyuFileInfo::CalculateIwyuViolations(vector<OneUse>* uses) {
       Union(associated_direct_includes, direct_includes());
 
   // (C2) + (C3) Find the minimal 'set cover' for all symbol uses.
+  const set<string>& associated_desired_includes = AssociatedDesiredIncludes();
   const set<string> desired_set_cover = internal::CalculateMinimalIncludes(
-      quoted_file_, direct_includes(), associated_direct_includes, uses);
+      quoted_file_, direct_includes(), associated_desired_includes, uses);
 
   // (C4) Remove .cc files from desired-includes unless they're in actual-inc.
   for (const string& header_name : desired_set_cover) {
@@ -1666,7 +1714,7 @@ void IwyuFileInfo::CalculateIwyuViolations(vector<OneUse>* uses) {
   // NOTE: this depends on our associated headers having had their
   // iwyu analysis done before us.
   set<string> effective_desired_includes = desired_includes();
-  InsertAllInto(AssociatedDesiredIncludes(), &effective_desired_includes);
+  InsertAllInto(associated_desired_includes, &effective_desired_includes);
 
   // Now that we've figured out desired_includes, figure out iwyu violations.
   for (OneUse& use : *uses) {
@@ -1742,15 +1790,15 @@ void ClearDesiredForSurplusIncludesOrForwardDeclares(ContainerType& container) {
     typename ContainerType::iterator v = ++container.lower_bound(k->first);
     typename ContainerType::iterator vend = container.upper_bound(k->first);
     for (; v != vend; ++v) {
-      v->second->clear_desired();
+      if (!v->second->is_elaborated_type())
+        v->second->clear_desired();
     }
   }
 }
 
 void CalculateDesiredIncludesAndForwardDeclares(
-    const vector<OneUse>& uses,
-    const set<string>& associated_desired_includes,
-    const set<const FileEntry*>& kept_includes,
+    const vector<OneUse>& uses, const set<string>& associated_desired_includes,
+    const set<OptionalFileEntryRef>& kept_includes,
     vector<OneIncludeOrForwardDeclareLine>* lines) {
   // First make sure all uses' includes and fwd decls are reflected in lines.
   for (const OneUse& use : uses) {
@@ -1865,9 +1913,22 @@ void CalculateDesiredIncludesAndForwardDeclares(
       line.set_desired();
     }
   }
+
+  // Keep all forward-decls fully declared inside a macro. Otherwise IWYU will
+  // suggest to remove the expansion of a macro which might do any number of
+  // other things with side effects.
+  for (OneIncludeOrForwardDeclareLine& line : *lines) {
+    if (!line.is_desired() && !line.IsIncludeLine() &&
+        IsDeclaredInsideMacro(line.fwd_decl())) {
+      VERRS(6) << "Retaining unused forward-declare inside macro: "
+               << GetFilePath(line.fwd_decl()) << ":" << line.LineNumberString()
+               << ": " << line.line() << "\n";
+      line.set_desired();
+    }
+  }
 }
 
-bool IsRemovablePrefixHeader(const FileEntry* file_entry,
+bool IsRemovablePrefixHeader(OptionalFileEntryRef file_entry,
                              const IwyuPreprocessorInfo* preprocessor_info) {
   if (file_entry) {
     IwyuFileInfo* file_info = preprocessor_info->FileInfoFor(file_entry);
@@ -1891,7 +1952,7 @@ void CleanupPrefixHeaderIncludes(
     if (line.is_present() && (policy == CommandlineFlags::kKeep))
       continue;  // Keep present line according to policy.
 
-    const FileEntry* file_entry = nullptr;
+    OptionalFileEntryRef file_entry;
     if (line.IsIncludeLine()) {
       file_entry = line.included_file();
       if (!file_entry)
@@ -1908,7 +1969,7 @@ void CleanupPrefixHeaderIncludes(
       line.clear_desired();
       VERRS(6) << "Ignoring '" << line.line()
                << "': is superseded by command line include "
-               << file_entry->getName() << "\n";
+               << GetFilePath(file_entry) << "\n";
     }
   }
 }
@@ -2093,7 +2154,7 @@ size_t PrintableDiffs(const string& filename,
       OutputLine("\nThe full include-list for " + filename + ":"));
     for (const auto& key_line : sorted_lines) {
       const OneIncludeOrForwardDeclareLine* line = key_line.second;
-      if (line->is_desired()) {
+      if (line->is_desired() && !line->is_elaborated_type()) {
         output_lines.push_back(
           PrintableIncludeOrForwardDeclareLine(*line, aqi));
       }
@@ -2137,16 +2198,16 @@ void IwyuFileInfo::HandlePreprocessingDone() {
     return;
   }
   bool should_report_violations = ShouldReportIWYUViolationsFor(file_);
-  std::list<const FileEntry*> direct_macro_use_includees;
+  std::list<OptionalFileEntryRef> direct_macro_use_includees;
   std::set_intersection(macro_users_.begin(), macro_users_.end(),
                         direct_includes_as_fileentries_.begin(),
                         direct_includes_as_fileentries_.end(),
                         std::inserter(direct_macro_use_includees,
                                       direct_macro_use_includees.end()));
-  for (const FileEntry* macro_use_includee : direct_macro_use_includees) {
+  for (OptionalFileEntryRef macro_use_includee : direct_macro_use_includees) {
     if (should_report_violations) {
-      ERRSYM(file_) << "Keep #include " << macro_use_includee->getName()
-                    << " in " << file_->getName()
+      ERRSYM(file_) << "Keep #include " << GetFilePath(macro_use_includee)
+                    << " in " << GetFilePath(file_)
                     << " because used macro is defined by includer.\n";
       ReportKnownDesiredFile(macro_use_includee);
     } else {
@@ -2239,4 +2300,22 @@ size_t IwyuFileInfo::CalculateAndReportIwyuViolations() {
   return num_edits;
 }
 
+bool IsDirectlyIncluded(const NamedDecl* decl, const IwyuFileInfo& includer) {
+  // If the decl or its header is private then a user should '#include' any of
+  // the specified public headers for it, so check their presence.
+  const vector<string> mapped_headers =
+      GlobalIncludePicker().GetMappedPublicHeaders(
+          GetWrittenQualifiedNameAsString(decl),
+          GetFilePath(includer.file_entry()), GetFilePath(decl));
+  if (!mapped_headers.empty()) {
+    auto pred = [&direct_includes =
+                     includer.direct_includes()](const string& header) {
+      return direct_includes.count(header);
+    };
+    return find_if(mapped_headers, pred) != mapped_headers.end();
+  }
+  // Otherwise, check the decl header directly.
+  return includer.direct_includes_as_fileentries().count(GetFileEntry(decl));
+}
+
 }  // namespace include_what_you_use
